@@ -65,6 +65,8 @@ static int rt_prio;
 
 static int use_tcp;
 
+static int keep;
+
 static unsigned int page_size;
 
 static const char *output_file = "trace.dat";
@@ -142,6 +144,19 @@ struct events {
 	char *name;
 };
 
+/* Files to be reset when done recording */
+struct reset_file {
+	struct reset_file	*next;
+	char			*path;
+	char			*reset;
+	int			prio;
+};
+
+static struct reset_file *reset_files;
+
+/* Triggers need to be cleared in a special way */
+static struct reset_file *reset_triggers;
+
 struct buffer_instance top_instance = { .keep = 1 };
 struct buffer_instance *buffer_instances;
 struct buffer_instance *first_instance = &top_instance;
@@ -165,6 +180,140 @@ static void init_instance(struct buffer_instance *instance)
 	instance->event_next = &instance->events;
 }
 
+enum {
+	RESET_DEFAULT_PRIO	= 0,
+	RESET_HIGH_PRIO		= 100000,
+};
+
+static void add_reset_file(const char *file, const char *val, int prio)
+{
+	struct reset_file *reset;
+	struct reset_file **last = &reset_files;
+
+	/* Only reset if we are not keeping the state */
+	if (keep)
+		return;
+
+	reset = malloc_or_die(sizeof(*reset));
+	reset->path = strdup(file);
+	reset->reset = strdup(val);
+	reset->prio = prio;
+	if (!reset->path || !reset->reset)
+		die("malloc");
+
+	while (*last && (*last)->prio > prio)
+		last = &(*last)->next;
+
+	reset->next = *last;
+	*last = reset;
+}
+
+static void add_reset_trigger(const char *file)
+{
+	struct reset_file *reset;
+
+	/* Only reset if we are not keeping the state */
+	if (keep)
+		return;
+
+	reset = malloc_or_die(sizeof(*reset));
+	reset->path = strdup(file);
+
+	reset->next = reset_triggers;
+	reset_triggers = reset;
+}
+
+static char *get_fd_content(int fd, const char *file)
+{
+	char *str = NULL;
+	int cnt = 0;
+	int ret;
+
+	for (;;) {
+		str = realloc(str, BUFSIZ * ++cnt);
+		if (!str)
+			die("malloc");
+		ret = read(fd, str + BUFSIZ * (cnt - 1), BUFSIZ);
+		if (ret < 0)
+			die("reading %s\n", file);
+		if (ret < BUFSIZ)
+			break;
+	}
+	str[BUFSIZ * (cnt-1) + ret] = 0;
+
+	return str;
+}
+
+static char *get_file_content(const char *file)
+{
+	char *str;
+	int fd;
+
+	fd = open(file, O_RDONLY);
+	if (fd < 0)
+		return NULL;
+
+	str = get_fd_content(fd, file);
+	close(fd);
+
+	return str;
+}
+
+/* To save the contents of the file */
+static void reset_save_file(const char *file, int prio)
+{
+	char *content;
+
+	content = get_file_content(file);
+	add_reset_file(file, content, prio);
+	free(content);
+}
+
+static char *strstrip(char *str)
+{
+	char *s;
+
+	if (!str)
+		return NULL;
+
+	s = str + strlen(str) - 1;
+	while (s >= str && isspace(*s))
+		s--;
+	s++;
+	*s = '\0';
+
+	for (s = str; *s && isspace(*s); s++)
+		;
+
+	return s;
+}
+
+/*
+ * @file: the file to check
+ * @nop: If the content of the file is this, use the reset value
+ * @reset: What to write if the file == @nop
+ */
+static void reset_save_file_cond(const char *file, int prio,
+				 const char *nop, const char *reset)
+{
+	char *content;
+	char *cond;
+
+	if (keep)
+		return;
+
+	content = get_file_content(file);
+
+	cond = strstrip(content);
+
+	if (strcmp(cond, nop) == 0)
+		add_reset_file(file, reset, prio);
+	else
+		add_reset_file(file, content, prio);
+
+	free(content);
+}
+
 /**
  * add_instance - add a buffer instance to the internal list
  * @instance: The buffer instance to add
@@ -420,6 +569,8 @@ static int set_ftrace_enable(const char *path, int set)
 	if (fd < 0)
 		return -ENODEV;
 
+	reset_save_file(path, RESET_DEFAULT_PRIO);
+
 	ret = -1;
 	fd = open(path, O_WRONLY);
 	if (fd < 0)
@@ -566,6 +717,7 @@ static void update_ftrace_pid(const char *pid, int reset)
 	static char *path;
 	int ret;
 	static int fd = -1;
+	static int first = 1;
 
 	if (!pid) {
 		if (fd >= 0)
@@ -588,6 +740,10 @@ static void update_ftrace_pid(const char *pid, int reset)
 			path = tracecmd_get_tracing_file("set_ftrace_pid");
 		if (!path)
 			return;
+		if (first) {
+			first = 0;
+			reset_save_file_cond(path, RESET_DEFAULT_PRIO, "no pid", "");
+		}
 		fd = open(path, O_WRONLY | O_CLOEXEC | (reset ? O_TRUNC : 0));
 		if (fd < 0)
 			return;
@@ -890,8 +1046,18 @@ set_plugin_instance(struct buffer_instance *instance, const char *name)
 
 	path = get_instance_file(instance, "current_tracer");
 	fp = fopen(path, "w");
-	if (!fp)
+	if (!fp) {
+		/*
+		 * Legacy kernels do not have current_tracer file, and they
+		 * always use nop. So, it doesn't need to try to change the
+		 * plugin for those if name is "nop".
+		 */
+		if (!strncmp(name, "nop", 3)) {
+			tracecmd_put_tracing_file(path);
+			return;
+		}
 		die("writing to '%s'", path);
+	}
 	tracecmd_put_tracing_file(path);
 
 	fwrite(name, 1, strlen(name), fp);
@@ -904,14 +1070,21 @@ set_plugin_instance(struct buffer_instance *instance, const char *name)
 	/* First try instance file, then top level */
 	path = get_instance_file(instance, "options/func_stack_trace");
 	fp = fopen(path, "w");
-	tracecmd_put_tracing_file(path);
 	if (!fp) {
+		tracecmd_put_tracing_file(path);
 		path = tracecmd_get_tracing_file("options/func_stack_trace");
 		fp = fopen(path, "w");
-		tracecmd_put_tracing_file(path);
-		if (!fp)
+		if (!fp) {
+			tracecmd_put_tracing_file(path);
 			return;
+		}
 	}
+	/*
+	 * Always reset func_stack_trace to zero. Don't bother saving
+	 * the original content.
+	 */
+	add_reset_file(path, "0", RESET_HIGH_PRIO);
+	tracecmd_put_tracing_file(path);
 	fwrite(&zero, 1, 1, fp);
 	fclose(fp);
 }
@@ -949,10 +1122,91 @@ static void set_option(const char *option)
 	fclose(fp);
 }
 
+static void add_reset_options(void)
+{
+	struct opt_list *opt;
+	const char *option;
+	char *content;
+	char *path;
+	char *ptr;
+	int len;
+
+	if (keep)
+		return;
+
+	path = tracecmd_get_tracing_file("trace_options");
+	content = get_file_content(path);
+
+	for (opt = options; opt; opt = opt->next) {
+		option = opt->option;
+		len = strlen(option);
+		ptr = content;
+ again:
+		ptr = strstr(ptr, option);
+		if (ptr) {
+			/* First make sure its the option we want */
+			if (ptr[len] != '\n') {
+				ptr += len;
+				goto again;
+			}
+			if (ptr - content >= 2 && strncmp(ptr - 2, "no", 2) == 0) {
+				/* Make sure this isn't ohno-option */
+				if (ptr > content + 2 && *(ptr - 3) != '\n') {
+					ptr += len;
+					goto again;
+				}
+				/* we enabled it */
+				ptr[len] = 0;
+				add_reset_file(path, ptr-2, RESET_DEFAULT_PRIO);
+				ptr[len] = '\n';
+				continue;
+			}
+			/* make sure this is our option */
+			if (ptr > content && *(ptr - 1) != '\n') {
+				ptr += len;
+				goto again;
+			}
+			/* this option hasn't changed, ignore it */
+			continue;
+		}
+
+		/* ptr is NULL, not found, maybe option is a no */
+		if (strncmp(option, "no", 2) != 0)
+			/* option is really not found? */
+			continue;
+
+		option += 2;
+		len = strlen(option);
+		ptr = content;
+ loop:
+		ptr = strstr(content, option);
+		if (!ptr)
+			/* Really not found? */
+			continue;
+
+		/* make sure this is our option */
+		if (ptr[len] != '\n') {
+			ptr += len;
+			goto loop;
+		}
+
+		if (ptr > content && *(ptr - 1) != '\n') {
+			ptr += len;
+			goto loop;
+		}
+
+		add_reset_file(path, option, RESET_DEFAULT_PRIO);
+	}
+	tracecmd_put_tracing_file(path);
+	free(content);
+}
+
 static void set_options(void)
 {
 	struct opt_list *opt;
 
+	add_reset_options();
+
 	while (options) {
 		opt = options;
 		options = opt->next;
@@ -1094,6 +1348,47 @@ static void write_file(const char *file, const char *str, const char *type)
 	}
 }
 
+enum {
+	STATE_NEWLINE,
+	STATE_SKIP,
+	STATE_COPY,
+};
+
+static int find_trigger(const char *file, char *buf, int size)
+{
+	FILE *fp;
+	int state = STATE_NEWLINE;
+	int ch;
+	int len = 0;
+
+	fp = fopen(file, "r");
+	while ((ch = fgetc(fp)) != EOF) {
+		if (ch == '\n') {
+			if (state == STATE_COPY)
+				break;
+			state = STATE_NEWLINE;
+			continue;
+		}
+		if (state == STATE_SKIP)
+			continue;
+		if (state == STATE_NEWLINE && ch == '#') {
+			state = STATE_SKIP;
+			continue;
+		}
+		if (state == STATE_COPY && ch == ':')
+			break;
+
+		state = STATE_COPY;
+		buf[len++] = ch;
+		if (len == size - 1)
+			break;
+	}
+	buf[len] = 0;
+	fclose(fp);
+
+	return len;
+}
+
 static void write_filter(const char *file, const char *filter)
 {
 	write_file(file, filter, "filter");
@@ -1104,6 +1399,53 @@ static void write_trigger(const char *file, const char *trigger)
 	write_file(file, trigger, "trigger");
 }
 
+static void clear_trigger(const char *file)
+{
+	char trigger[BUFSIZ];
+	int len;
+
+	trigger[0] = '!';
+
+	/*
+	 * To delete a trigger, we need to write a '!trigger'
+	 * to the file for each trigger.
+	 */
+	do {
+		len = find_trigger(file, trigger+1, BUFSIZ-1);
+		if (len)
+			write_trigger(file, trigger);
+	} while (len);
+}
+
+static void update_reset_triggers(void)
+{
+	struct reset_file *reset;
+
+	while (reset_triggers) {
+		reset = reset_triggers;
+		reset_triggers = reset->next;
+
+		clear_trigger(reset->path);
+		free(reset->path);
+		free(reset);
+	}
+}
+
+static void update_reset_files(void)
+{
+	struct reset_file *reset;
+
+	while (reset_files) {
+		reset = reset_files;
+		reset_files = reset->next;
+
+		write_file(reset->path, reset->reset, "reset");
+		free(reset->path);
+		free(reset->reset);
+		free(reset);
+	}
+}
+
 static void
 update_event(struct event_list *event, const char *filter,
 	     int filter_only, char update)
@@ -1120,10 +1462,15 @@ update_event(struct event_list *event, const char *filter,
 		return;
 	}
 
-	if (filter && event->filter_file)
+	if (filter && event->filter_file) {
+		reset_save_file_cond(event->filter_file, RESET_DEFAULT_PRIO,
+				     "none", "0");
 		write_filter(event->filter_file, filter);
+	}
 
 	if (event->trigger_file) {
+		add_reset_trigger(event->trigger_file);
+		clear_trigger(event->trigger_file);
 		write_trigger(event->trigger_file, event->trigger);
 		/* Make sure we don't write this again */
 		free(event->trigger_file);
@@ -1472,7 +1819,7 @@ static void make_sched_event(struct buffer_instance *instance,
 	if (*event)
 		return;
 
-	path = malloc_or_die(strlen(sched->filter_file) + strlen(sched_path) + 1);
+	path = malloc_or_die(strlen(sched->filter_file) + strlen(sched_path) + 2);
 
 	sprintf(path, "%s", sched->filter_file);
 
@@ -1481,6 +1828,7 @@ static void make_sched_event(struct buffer_instance *instance,
 	sprintf(p, "%s/filter", sched_path);
 
 	*event = create_event(instance, path, sched);
+	free(path);
 }
 
 static void test_event(struct event_list *event, const char *path,
@@ -1949,6 +2297,9 @@ static void start_threads(void)
 
 	memset(pids, 0, sizeof(*pids) * cpu_count * (buffers + 1));
 
+	/* Make sure all output is flushed before forking */
+	fflush(stdout);
+
 	for_all_instances(instance) {
 		int x;
 		for (x = 0; x < cpu_count; x++)
@@ -1995,7 +2346,7 @@ static void touch_file(const char *file)
 {
 	int fd;
 
-	fd = open(file, O_WRONLY | O_CREAT | O_TRUNC);
+	fd = open(file, O_WRONLY | O_CREAT | O_TRUNC, 0644);
 	if (fd < 0)
 		die("could not create file %s\n", file);
 	close(fd);
@@ -2241,7 +2592,7 @@ static unsigned long long find_time_stamp(struct pevent *pevent)
 			continue;
 
 		sprintf(file, "%s/%s/trace_pipe_raw", path, name);
-		fd = open(file, O_RDONLY);
+		fd = open(file, O_RDONLY | O_NONBLOCK);
 		if (fd < 0)
 			continue;
 		do {
@@ -2493,6 +2844,7 @@ static void remove_instances(void)
 static void check_plugin(const char *plugin)
 {
 	char *buf;
+	char *str;
 	char *tok;
 
 	/*
@@ -2506,8 +2858,9 @@ static void check_plugin(const char *plugin)
 	if (!buf)
 		die("No plugins available");
 
-	while ((tok = strtok(buf, " "))) {
-		buf = NULL;
+	str = buf;
+	while ((tok = strtok(str, " "))) {
+		str = NULL;
 		if (strcmp(tok, plugin) == 0)
 			goto out;
 	}
@@ -2692,7 +3045,6 @@ void trace_record (int argc, char **argv)
 	int extract = 0;
 	int run_command = 0;
 	int neg_event = 0;
-	int keep = 0;
 	int date = 0;
 
 	int c;
@@ -3170,6 +3522,9 @@ void trace_record (int argc, char **argv)
 	if (keep)
 		exit(0);
 
+	update_reset_files();
+	update_reset_triggers();
+
 	set_plugin("nop");
 
 	remove_instances();
